import os

import six
from django.test import TransactionTestCase
from django.http.request import MultiValueDict
from django.core.files.uploadedfile import SimpleUploadedFile

from ..backend import utils
from ..forms import (
    factory,
    WooeyForm,
)
from ..forms import config as forms_config
from ..models import ScriptVersion, WooeyJob

from . import (
    config,
    factories,
    mixins,
    utils as test_utils,
)


class FormTestCase(mixins.ScriptFactoryMixin, mixins.FileCleanupMixin, TransactionTestCase):

    @staticmethod
    def get_mvdict(data):
        return MultiValueDict(data)

    def test_master_form(self):
        script_version = self.translate_script
        form = utils.get_master_form(script_version=script_version)
        self.assertTrue(isinstance(form, WooeyForm))
        qdict = self.get_mvdict(config.SCRIPT_DATA['translate'].get('data'))
        utils.validate_form(form=form, data=qdict,
                            files=config.SCRIPT_DATA['translate'].get('files'))
        self.assertTrue(form.is_valid())

    def test_groups_dont_duplicate(self):
        # fixes https://github.com/wooey/Wooey/issues/163
        # Update the translate script and make sure we do not duplicate form groups
        # Make a new script version
        old_translate_pk = self.translate_script.pk
        new_translate = self.translate_script
        new_translate.pk = None
        new_translate.save()
        old_translate = ScriptVersion.objects.get(pk=old_translate_pk)

        # Use a script that is slightly different but the same form groups
        script_path = test_utils.save_script_path(os.path.join(config.WOOEY_TEST_SCRIPTS, 'translate2.py'))
        new_translate = utils.add_wooey_script(
            script_version=new_translate,
            script_path=script_path,
        )['script']

        # Assert we updated correctly
        self.assertEqual(old_translate.script.pk, new_translate.script.pk)
        self.assertEqual(old_translate.script_iteration, 1)
        self.assertEqual(new_translate.script_iteration, 2)

        # Make sure we still have the same number of groups after updating
        new_form = utils.get_form_groups(script_version=new_translate)
        old_form = utils.get_form_groups(script_version=old_translate)
        for new_parsers, old_parsers in zip(new_form['parsers'], old_form['parsers']):
            for new_subparser, old_subparser in zip(new_form['parsers'][new_parsers], old_form['parsers'][old_parsers]):
                self.assertEqual(new_subparser['group_name'], old_subparser['group_name'])

            self.assertEqual(len(new_parsers), len(old_parsers))

        self.assertEqual(len(new_form['parsers']), len(old_form['parsers']))

    def test_group_form(self):
        script_version = self.translate_script
        form = utils.get_form_groups(script_version=script_version)
        subparser = script_version.scriptparser_set.first()
        subparser_key = (subparser.pk, subparser.name)
        self.assertEqual(len(form['parsers'][subparser_key]), 2)
        # test we can hide parameters and groups
        from wooey.models import ScriptParameterGroup, ScriptParameter
        groups = ScriptParameterGroup.objects.filter(script_version=script_version)
        group = groups[1]
        group.hidden = True
        group.save()
        form = utils.get_form_groups(script_version=script_version)
        self.assertEqual(len(form['parsers'][subparser_key]), 1, 'Script Parameter group is hidden but shown')
        group.hidden = False
        group.save()
        form = utils.get_form_groups(script_version=script_version)
        self.assertEqual(len(form['parsers'][subparser_key]), 2, 'Script Parameter group is shown but hidden')
        param = ScriptParameter.objects.get(script_version=script_version, slug='out')
        param.hidden = True
        param.save()
        form = utils.get_form_groups(script_version=script_version)
        slug = test_utils.get_subparser_form_slug(script_version, 'out')
        self.assertNotIn(slug, form['parsers'][subparser_key][1]['form'].fields, 'Script Parameter is hidden but shown')
        param.hidden = False
        param.save()
        form = utils.get_form_groups(script_version=script_version)
        self.assertIn(slug, form['parsers'][subparser_key][1]['form'].fields, 'Script Parameter is shown but hidden')


    def test_multiplechoice_form(self):
        script_version = self.choice_script
        form = utils.get_master_form(script_version=script_version)
        # check our wrapper is in the form render
        form_str = str(form)
        self.assertTrue(forms_config.WOOEY_MULTI_WIDGET_ANCHOR in form_str)
        self.assertTrue(forms_config.WOOEY_MULTI_WIDGET_ATTR in form_str)

        qdict = {}
        for key,value in six.iteritems(config.SCRIPT_DATA['choices'].get('data')):
            try:
                form_slug = test_utils.get_subparser_form_slug(script_version, key)
                qdict[form_slug] = value
            except:
                qdict[key] = value
        qdict = self.get_mvdict(qdict)

        fdict = {}
        for key, value in six.iteritems(config.SCRIPT_DATA['choices'].get('files')):
            try:
                form_slug = test_utils.get_subparser_form_slug(script_version, key)
                fdict[form_slug] = value
            except:
                fdict[key] = value

        uploaded_files = {}
        storage = utils.get_storage(local=True)
        for file_key, files in fdict.items():
            uploaded_files[file_key] = []
            for file_location in files:
                file_name = os.path.split(file_location.name)[1]
                uploaded_files[file_key].append(SimpleUploadedFile(file_name, file_location.read()))

        utils.validate_form(
            form=form,
            data=qdict,
            files=uploaded_files
        )
        self.assertTrue(form.is_valid())
        # test we can create a job from this form
        # this is implemented to put data and files in the same dictionary, so update it
        form.cleaned_data.update(uploaded_files)
        job = utils.create_wooey_job(
            script_version_pk=script_version.pk,
            data=form.cleaned_data
        )
        # check the files are here
        file_param = test_utils.get_subparser_form_slug(script_version, 'multiple_file_choices')
        files = [i.value for i in job.get_parameters() if i.parameter.form_slug == file_param]
        self.assertEqual(len(files), len(uploaded_files.get(file_param)))

    def test_without_args_form(self):
        script_version = self.without_args
        form = utils.get_master_form(script_version=script_version)
        # check our wrapper is in the form render
        # in a normal POST request, the data will contain {'wooey_type': script_pk}. This will cause the is_bound
        # attribute of the form to be set. Otherwise, this test will always fail. It is OK to set this data
        # in the test, since it is how it will be recapitulated in the normal use case
        form_data = form.initial
        form_data.update(config.SCRIPT_DATA['without_args'].get('data'))
        utils.validate_form(form=form, data=form_data)
        self.assertTrue(form.is_valid(), form.errors)
        # test we can create a job from this form
        job = utils.create_wooey_job(script_version_pk=script_version.pk, data=form.cleaned_data)
        job.submit_to_celery()
        # get the job again
        job = WooeyJob.objects.get(pk=job.pk)
        self.assertEqual(job.status, WooeyJob.COMPLETED)

    def test_wooey_form(self):
        # Make sure wooey form exists and has everything we need in it
        script_version = self.without_args
        forms = utils.get_form_groups(script_version=self.without_args)
        wooey_form = forms['wooey_form']
        self.assertDictEqual(wooey_form.initial, {'wooey_type': script_version.pk})

    def test_multiple_file_field_initial_value(self):
        # Addresses https://github.com/wooey/Wooey/issues/248
        script_version = self.choice_script
        multiple_files_param = script_version.scriptparameter_set.get(script_param='multiple_file_choices')
        # Upload some fake files
        storage = utils.get_storage(local=False)
        storage.save('file1', SimpleUploadedFile('file1', b'abc'))
        file1_path = storage.path('file1')
        storage.save('file2', SimpleUploadedFile('file2', b'abc'))
        file2_path = storage.path('file2')
        form = utils.get_form_groups(script_version=script_version, initial_dict={
            multiple_files_param.form_slug: ['file1', 'file2']
        })
        # TODO: Make a function to ease this
        initial_files = [i.path for i in form['parsers'][(multiple_files_param.parser.pk, multiple_files_param.parser.name)][1]['form'].fields[multiple_files_param.form_slug].initial]
        self.assertIn(file1_path, initial_files)
        self.assertIn(file2_path, initial_files)

    def test_form_with_custom_widget(self):
        script_version = self.choice_script
        # Associate a custom widget with a field
        choice_param = script_version.scriptparameter_set.get(script_param='one_choice')
        widget = factories.WooeyWidgetFactory(
            widget_class='django.forms.TextInput',
            input_class='custom',
            input_properties='custom-property',
            input_attributes='attr1="custom1" attr2="custom2"'
        )
        choice_param.custom_widget = widget
        choice_param.save()
        form = utils.get_form_groups(script_version=script_version)

        from django.forms import TextInput
        field = form['parsers'][(choice_param.parser.pk, choice_param.parser.name)][1]['form'].fields[choice_param.form_slug]
        self.assertTrue(isinstance(field.widget, TextInput))
        self.assertEquals(
            field.widget.attrs,
            {
                'custom-property': True,
                'attr1': 'custom1',
                'attr2': 'custom2',
                'class': 'custom',
            }
        )

    def test_handles_special_characters(self):
        script_version = self.choice_script
        # Associate a custom widget with a field
        choice_str_slug = test_utils.get_subparser_form_slug(script_version, 'choices_str')
        choice_str = 'This & is special'
        form = utils.get_master_form(script_version=script_version)
        qdict = self.get_mvdict({choice_str_slug: [choice_str], 'wooey_type': [script_version.pk], 'job_name': ['abc']})
        utils.validate_form(form=form, data=qdict)
        self.assertTrue(form.is_valid(), form.errors)
        self.assertEqual(form.cleaned_data[choice_str_slug], [choice_str])

    def test_subparser_selection(self):
        script_version = self.subparser_script
        subparser = script_version.scriptparser_set.get(name='subparser2')
        form = utils.get_master_form(script_version=script_version, parser=subparser.pk)
        sp1_slug = test_utils.get_subparser_form_slug(script_version, 'sp1')
        sp2_slug = test_utils.get_subparser_form_slug(script_version, 'sp2')
        test_arg_slug = test_utils.get_subparser_form_slug(script_version, 'test_arg')

        # The sp1 slug is only present in subparser1, so we should not see it, but we should
        # see the sp2 slug and the general subparser options of test_arg
        self.assertIn(sp2_slug, form.fields)
        self.assertNotIn(sp1_slug, form.fields)
        self.assertIn(test_arg_slug, form.fields)


import argparse
import sys

parser = argparse.ArgumentParser(description="Something")
parser.add_argument("--output_filename", type=argparse.FileType('w'), required=True, help="Name of Excel output file.", nargs='+')
parser.add_argument("--due_date_field", type=str, help="Column name for due date field. 'Due Date' by default.")

if __name__ == '__main__':
    args = parser.parse_args()



from __future__ import absolute_import
import os

from celery import Celery


os.environ.setdefault('DJANGO_SETTINGS_MODULE', '{{ project_name }}.settings')

app = Celery('{{ project_name }}')

# Using a string here means the worker will not have to
# pickle the object when using Windows.
app.config_from_object('django.conf:settings', namespace='CELERY')
app.autodiscover_tasks()


@app.task(bind=True)
def debug_task(self):
    print('Request: {0!r}'.format(self.request))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2019-06-09 08:15
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('wooey', '0040_add-widget-class'),
    ]

    operations = [
        migrations.AlterField(
            model_name='scriptparameter',
            name='parser',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='wooey.ScriptParser'),
        ),
        migrations.AlterField(
            model_name='scriptparameter',
            name='script_param',
            field=models.TextField(),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('wooey', '0010_script_versioning_data_migration'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='scriptparameter',
            name='script',
        ),
        migrations.RemoveField(
            model_name='scriptparametergroup',
            name='script',
        ),
        migrations.RemoveField(
            model_name='wooeyjob',
            name='script',
        ),
        migrations.RemoveField(
            model_name='script',
            name='script_path',
        ),
        migrations.RemoveField(
            model_name='script',
            name='script_version',
        ),
        migrations.AlterField(
            model_name='scriptversion',
            name='script',
            field=models.ForeignKey(related_name='script_version', to='wooey.Script', on_delete=models.PROTECT),
        ),
        migrations.AlterField(
            model_name='scriptparameter',
            name='script_version',
            field=models.ForeignKey(to='wooey.ScriptVersion', on_delete=models.PROTECT),
        ),
        migrations.AlterField(
            model_name='scriptparametergroup',
            name='script_version',
            field=models.ForeignKey(to='wooey.ScriptVersion', on_delete=models.PROTECT),
        ),
        migrations.AlterField(
            model_name='wooeyjob',
            name='script_version',
            field=models.ForeignKey(to='wooey.ScriptVersion', on_delete=models.PROTECT),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import uuid


def gen_uuid(apps, schema_editor):
    WooeyJob = apps.get_model('wooey', 'WooeyJob')
    for obj in WooeyJob.objects.all():
        obj.uuid = uuid.uuid4()
        obj.save()


class Migration(migrations.Migration):

    dependencies = [
        ('wooey', '0012_wooeyjob_uuid'),
    ]

    operations = [
        # Set the uuids for existing records
        migrations.RunPython(gen_uuid),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wooey', '0014_wooeyjob_uuid_finalise'),
    ]

    operations = [
        migrations.AddField(
            model_name='scriptparameter',
            name='hidden',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='scriptparametergroup',
            name='hidden',
            field=models.BooleanField(default=False),
        ),
    ]



from . version import DJANGO_VERSION, DJ18, DJ17, DJ16, DJ19, DJ110, DJ111, DJ20, DJ21
from django.conf import settings

try:
    settings.configure()
except RuntimeError:
    pass

if DJANGO_VERSION < DJ19:
    from django.template.base import TagHelperNode, parse_bits
else:
    from django.template.library import TagHelperNode, parse_bits

if DJANGO_VERSION >= DJ18:
    from django.template import Engine
    get_template_from_string = Engine.get_default().from_string
else:
    from django.template.loader import get_template_from_string

if DJANGO_VERSION >= DJ20:
    from django.urls import reverse
else:
    from django.core.urlresolvers import reverse



#!/usr/bin/env python

__author__ = 'chris'
import argparse
import os
import sys
import pandas as pd
import seaborn as sns
import numpy as np

parser = argparse.ArgumentParser(description="Create a heatmap from a delimited file.")
parser.add_argument('--tsv', help='The delimited file to plot.', type=argparse.FileType('r'), required=True)
parser.add_argument('--delimiter', help='The delimiter for fields. Default: tab', type=str, default='\t')
parser.add_argument('--row', help='The column containing row to create a heatmap from. Default to first row.', type=str)
parser.add_argument('--cols', help='The columns to choose values from (separate by a comma for multiple). Default: All non-rows', type=str)
parser.add_argument('--log-normalize', help='Whether to log normalize data.', action='store_true')

def main():
    args = parser.parse_args()
    data = pd.read_table(args.tsv, index_col=args.row if args.row else 0, sep=args.delimiter, encoding='utf-8')
    if args.cols:
        try:
            data = data.loc[:,args.cols.split(',')]
        except KeyError:
            data = data.iloc[:,[int(i)-1 for i in args.cols.split(',')]]
    if len(data.columns) > 50:
        raise BaseException('Too many columns')
    data = np.log2(data) if args.log_normalize else data
    data[data==-1*np.inf] = data[data!=-1*np.inf].min().min()
    width = 5+0 if len(data.columns)<50 else (len(data.columns)-50)/100
    row_cutoff = 1000
    height = 15+0 if len(data)<row_cutoff else (len(data)-row_cutoff)/75.0
    seaborn_map = sns.clustermap(data, figsize=(width, height))
    seaborn_map.savefig('{}_heatmap.png'.format(os.path.split(args.tsv.name)[1]))
    seaborn_map.data2d.to_csv('{}_heatmap.tsv'.format(os.path.split(args.tsv.name)[1]), sep='\t')

if __name__ == "__main__":
    sys.exit(main())



from .django_urls import *
from django.conf.urls import include, url

urlpatterns += [
    #url(r'^admin/', admin.site.urls),
    url(r'^', include('wooey.urls')),
    url(r'^', include('django.contrib.auth.urls')),
]



#!/usr/bin/env python

__author__ = 'Chris Mitchell'

import argparse
import os
import socket
import sys
from urllib import FancyURLopener
from apiclient import discovery

description = """
This will find you cats, and optionally, kitties.
"""

socket.setdefaulttimeout(10)

parser = argparse.ArgumentParser(description = description)
parser.add_argument('--count', help='The number of cats to find (max: 10)', type=int, default=1)
parser.add_argument('--kittens', help='Search for kittens.', action='store_true')
parser.add_argument('--breed', help='The breed of cat to find', type=str, choices=('lol', 'tabby', 'bengal', 'scottish', 'grumpy'))

# Start FancyURLopener with defined version
class MyOpener(FancyURLopener):
    version = 'Mozilla/5.0 (Windows; U; Windows NT 5.1; it; rv:1.8.1.11)Gecko/20071127     Firefox/2.0.0.11'

myopener = MyOpener()

def main():
    args = parser.parse_args()
    searchTerm = 'kittens' if args.kittens else 'cats'
    cat_count = args.count if args.count < 10 else 10
    if args.breed:
        searchTerm += ' {0}'.format(args.breed)

    # Notice that the start changes for each iteration in order to request a new set of images for each loop
    service = discovery.build('customsearch', 'v1', developerKey=os.environ.get('GOOGLE_DEV_KEY'))
    cse = service.cse()
    search_kwrds = {
        'q': searchTerm,
        'cx': os.environ.get('GOOGLE_CX'),
        'fileType': 'jpg',
        'imgType': 'photo',
        'num': cat_count,
        'searchType': 'image'
    }
    request = cse.list(**search_kwrds)
    response = request.execute()
    for item in response.get('items', []):
        url = item.get('link')
        filename = url.split('/')[-1]
        try:
            myopener.retrieve(url, filename)
        except IOError:
            continue


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('wooey', '0003_populate_from_slug'),
    ]

    operations = [
        migrations.AlterField(
            model_name='scriptparameter',
            name='choice_limit',
            field=models.CharField(max_length=10, null=True, blank=True),
        ),
    ]



from __future__ import absolute_import

import os

from django.core.files.storage import get_storage_class, FileSystemStorage
from storages.backends.s3boto3 import S3Boto3Storage

from . import settings as wooey_settings

class CachedS3Boto3Storage(S3Boto3Storage):
    def __init__(self, *args, **kwargs):
        if os.environ.get('TESTING', False):
            from .tests import config
            kwargs['location'] = config.WOOEY_TEST_REMOTE_STORAGE_DIR
        super(CachedS3Boto3Storage, self).__init__(*args, **kwargs)
        self.local_storage = get_storage_class('django.core.files.storage.FileSystemStorage')()

    def _open(self, name, mode='rb'):
        original_file = super(CachedS3Boto3Storage, self)._open(name, mode=mode)
        if name.endswith('.gz'):
            return original_file
        return original_file

    def path(self, name):
        return self.local_storage.path(name)

    def delete(self, name):
        # we have to remove the name from the _entries cache or else deleted files will persist in our cache
        # and give false information
        super(CachedS3Boto3Storage, self).delete(name)
        name = self._normalize_name(self._clean_name(name))
        encoded_name = self._encode_name(name)
        self._entries.pop(encoded_name, None)


class FakeRemoteStorage(FileSystemStorage):
    def __init__(self, *args, **kwargs):
        from .tests import config
        kwargs['location'] = config.WOOEY_TEST_REMOTE_STORAGE_PATH
        super(FakeRemoteStorage, self).__init__(*args, **kwargs)
        self.local_storage = get_storage_class('django.core.files.storage.FileSystemStorage')()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wooey', '0015_hidden_parameters'),
    ]

    operations = [
        migrations.AddField(
            model_name='wooeyfile',
            name='checksum',
            field=models.CharField(max_length=40, blank=True),
        ),
    ]



from __future__ import absolute_import, print_function, unicode_literals
import os
import importlib
import json
import six
import uuid
from io import IOBase

from autoslug import AutoSlugField
from celery import states
from django.db import models, transaction
from django.conf import settings
from django.core.cache import caches as django_cache
from django.core.files.storage import SuspiciousFileOperation
from django.contrib.auth.models import Group
from django.utils.translation import ugettext_lazy as _
from django.db import transaction
from django.utils.text import get_valid_filename
from jsonfield import JSONCharField

from ..django_compat import reverse
from . mixins import UpdateScriptsMixin, ModelDiffMixin, WooeyPy2Mixin
from .. import settings as wooey_settings
from .. backend import utils

# TODO: Handle cases where celery is not setup but specified to be used
tasks = importlib.import_module(wooey_settings.WOOEY_CELERY_TASKS)


class ScriptGroup(UpdateScriptsMixin, WooeyPy2Mixin, models.Model):
    """
        This is a group of scripts, it holds general information
        about a collection of scripts, and allows for custom descriptions

    """
    group_name = models.TextField()
    slug = AutoSlugField(populate_from='group_name', unique=True)
    group_description = models.TextField(null=True, blank=True)
    group_order = models.SmallIntegerField(default=1)
    is_active = models.BooleanField(default=True)
    user_groups = models.ManyToManyField(Group, blank=True)

    class Meta:
        app_label = 'wooey'
        verbose_name = _('script group')
        verbose_name_plural = _('script groups')

    def __str__(self):
        return self.group_name


class Script(ModelDiffMixin, WooeyPy2Mixin, models.Model):
    script_name = models.CharField(max_length=255)
    slug = AutoSlugField(populate_from='script_name', unique=True)
    # we create defaults for the script_group in the clean method of the model. We have to set it to null/blank=True
    # or else we will fail form validation before we hit the model.
    script_group = models.ForeignKey('ScriptGroup', null=True, blank=True, on_delete=models.PROTECT)
    script_description = models.TextField(blank=True, null=True)
    documentation = models.TextField(blank=True, null=True)
    script_order = models.PositiveSmallIntegerField(default=1)
    is_active = models.BooleanField(default=True)
    user_groups = models.ManyToManyField(Group, blank=True)

    execute_full_path = models.BooleanField(default=True)  # use full path for subprocess calls
    save_path = models.CharField(max_length=255, blank=True, null=True,
                                 help_text='By default save to the script name,'
                                           ' this will change the output folder.')

    created_date = models.DateTimeField(auto_now_add=True)
    modified_date = models.DateTimeField(auto_now=True)

    class Meta:
        app_label = 'wooey'
        verbose_name = _('script')
        verbose_name_plural = _('scripts')

    def __str__(self):
        return self.script_name

    def get_url(self):
        return reverse('wooey:wooey_script', kwargs={'slug': self.slug})

    @property
    def latest_version(self):
        return self.script_version.get(default_version=True)

    def clean(self):
        if self.script_group is None:
            group = ScriptGroup.objects.filter(group_name=wooey_settings.WOOEY_DEFAULT_SCRIPT_GROUP).order_by('pk').first()
            if not group:
                group, created = ScriptGroup.objects.get_or_create(group_name=wooey_settings.WOOEY_DEFAULT_SCRIPT_GROUP)
            self.script_group = group

    def get_previous_versions(self):
        return self.script_version.all().order_by('script_version', 'script_iteration')


class ScriptVersion(ModelDiffMixin, WooeyPy2Mixin, models.Model):
    # when a script updates, increment this to keep old scripts that are cloned working. The downside is we get redundant
    # parameters, but even a huge site may only have a few thousand parameters to query though.
    script_version = models.CharField(max_length=50, help_text='The script version.', blank=True, default='1')
    script_iteration = models.PositiveSmallIntegerField(default=1)
    script_path = models.FileField()
    default_version = models.BooleanField(default=False)
    script = models.ForeignKey('Script', related_name='script_version', on_delete=models.CASCADE)
    checksum = models.CharField(max_length=40, blank=True)

    created_date = models.DateTimeField(auto_now_add=True)
    modified_date = models.DateTimeField(auto_now=True)

    error_messages = {
        'duplicate_script': _('This script already exists!'),
    }

    class Meta:
        app_label = 'wooey'
        verbose_name = _('script version')
        verbose_name_plural = _('script versions')

    def __str__(self):
        return '{}({}: {})'.format(self.script.script_name, self.script_version, self.script_iteration)

    def get_url(self):
        return reverse('wooey:wooey_script', kwargs={'slug': self.script.slug})

    def get_version_url(self):
        return reverse('wooey:wooey_script', kwargs={'slug': self.script.slug, 'script_version': self.script_version, 'script_iteration': self.script_iteration})

    def get_script_path(self):
        local_storage = utils.get_storage(local=True)
        path = local_storage.path(self.script_path.path)
        return path if self.script.execute_full_path else os.path.split(path)[1]

    def get_parameters(self):
        return ScriptParameter.objects.filter(script_version=self).order_by('param_order', 'pk')


class WooeyJob(WooeyPy2Mixin, models.Model):
    """
    This model serves to link the submitted celery tasks to a script submitted
    """
    # blank=True, null=True is to allow anonymous users to submit jobs
    user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=models.SET_NULL)
    celery_id = models.CharField(max_length=255, null=True)
    uuid = models.CharField(max_length=255, default=uuid.uuid4, unique=True)
    job_name = models.CharField(max_length=255)
    job_description = models.TextField(null=True, blank=True)
    stdout = models.TextField(null=True, blank=True)
    stderr = models.TextField(null=True, blank=True)

    COMPLETED = 'completed'
    DELETED = 'deleted'
    FAILED = states.FAILURE
    RUNNING = 'running'
    SUBMITTED = 'submitted'

    STATUS_CHOICES = (
        (COMPLETED, _('Completed')),
        (DELETED, _('Deleted')),
        (FAILED, _('Failed')),
        (RUNNING, _('Running')),
        (SUBMITTED, _('Submitted')),
    )

    status = models.CharField(max_length=255, default=SUBMITTED, choices=STATUS_CHOICES)

    save_path = models.CharField(max_length=255, blank=True, null=True)
    command = models.TextField()
    created_date = models.DateTimeField(auto_now_add=True)
    modified_date = models.DateTimeField(auto_now=True)
    script_version = models.ForeignKey('ScriptVersion', on_delete=models.PROTECT)

    error_messages = {
        'invalid_permissions': _('You are not authenticated to view this job.'),
    }

    class Meta:
        app_label = 'wooey'
        verbose_name = _('wooey job')
        verbose_name_plural = _('wooey jobs')

    def __str__(self):
        return self.job_name

    def get_parameters(self):
        return ScriptParameters.objects.select_related('parameter').filter(job=self).order_by('pk')

    def submit_to_celery(self, **kwargs):
        if kwargs.get('resubmit'):
            params = self.get_parameters()
            user = kwargs.get('user')
            self.pk = None
            self.user = None if user is None or not user.is_authenticated else user
            # clear the output channels
            self.celery_id = None
            self.uuid = uuid.uuid4()
            self.stdout = ''
            self.stderr = ''
            self.save()
            with transaction.atomic():
                for param in params:
                    param.pk = None
                    param.job = self
                    param.recreate()
                    param.save()
        self.status = self.SUBMITTED
        self.save()
        task_kwargs = {'wooey_job': self.pk, 'rerun': kwargs.pop('rerun', False)}

        if task_kwargs.get('rerun'):
            utils.purge_output(job=self)
        if wooey_settings.WOOEY_CELERY:
            transaction.on_commit(lambda: tasks.submit_script.delay(**task_kwargs))
        else:
            transaction.on_commit(lambda: tasks.submit_script(**task_kwargs))
        return self

    def get_resubmit_url(self):
        return reverse('wooey:wooey_script_clone', kwargs={'slug': self.script_version.script.slug, 'job_id': self.pk})

    @property
    def output_path(self):
        return os.path.join(wooey_settings.WOOEY_FILE_DIR,
                            get_valid_filename(self.user.username if self.user is not None else ''),
                            get_valid_filename(self.script_version.script.slug if not self.script_version.script.save_path else self.script_version.script.save_path),
                            str(self.uuid))

    def get_output_path(self):
        path = self.output_path
        utils.mkdirs(os.path.join(settings.MEDIA_ROOT, path))
        return path

    def get_upload_path(self):
        path = self.output_path
        utils.mkdirs(os.path.join(settings.MEDIA_ROOT, path))
        return path

    def get_relative_path(self, path):
        # We make the filename relative to the MEDIA_ROOT, this is for filesystems that can change between
        # machines. We also want to omit any leading path separators so we can join the path to whatever
        # MEDIA_ROOT is currently at work instead of giving a path from a root
        return path[path.find(self.get_output_path()):].lstrip(os.path.sep)

    def get_realtime_key(self):
        return 'wooeyjob_{}_rt'.format(self.pk)

    def update_realtime(self, stdout='', stderr='', delete=False):
        wooey_cache = wooey_settings.WOOEY_REALTIME_CACHE
        if delete == False and wooey_cache is None:
            self.stdout = stdout
            self.stderr = stderr
            self.save()
        elif wooey_cache is not None:
            cache = django_cache[wooey_cache]
            if delete:
                cache.delete(self.get_realtime_key())
            else:
                cache.set(self.get_realtime_key(), json.dumps({'stdout': stdout, 'stderr': stderr}))

    def get_realtime(self):
        wooey_cache = wooey_settings.WOOEY_REALTIME_CACHE
        if wooey_cache is not None:
            cache = django_cache[wooey_cache]
            out = cache.get(self.get_realtime_key())
            if out:
                return json.loads(out)
        return {'stdout': self.stdout, 'stderr': self.stderr}

    def get_stdout(self):
        if self.status != WooeyJob.COMPLETED:
            rt = self.get_realtime().get('stdout')
            if rt:
                return rt
        return self.stdout

    def get_stderr(self):
        if self.status != WooeyJob.COMPLETED:
            rt = self.get_realtime().get('stderr')
            if rt:
                return rt
        return self.stderr


class ScriptParameterGroup(UpdateScriptsMixin, WooeyPy2Mixin, models.Model):
    group_name = models.TextField()
    hidden = models.BooleanField(default=False)
    script_version = models.ManyToManyField('ScriptVersion')

    class Meta:
        app_label = 'wooey'
        verbose_name = _('script parameter group')
        verbose_name_plural = _('script parameter groups')

    def __str__(self):
        return '{}: {}'.format(self.script_version.first().script.script_name, self.group_name)


class ScriptParser(WooeyPy2Mixin, models.Model):
    name = models.CharField(max_length=255, blank=True, default='')
    script_version = models.ManyToManyField('ScriptVersion')

    def __str__(self):
        script_version = self.script_version.first()
        return '{}: {}'.format(script_version.script.script_name if script_version else 'No Script Assigned', self.name)


class ScriptParameter(UpdateScriptsMixin, WooeyPy2Mixin, models.Model):
    """
        This holds the parameter mapping for each script, and enforces uniqueness by each script via a FK.
    """
    parser = models.ForeignKey('ScriptParser', on_delete=models.PROTECT)
    script_version = models.ManyToManyField('ScriptVersion')
    short_param = models.CharField(max_length=255, blank=True)
    script_param = models.TextField()
    slug = AutoSlugField(populate_from='script_param', unique=True)
    is_output = models.BooleanField(default=None)
    required = models.BooleanField(default=False)
    choices = models.CharField(max_length=255, null=True, blank=True)
    choice_limit = models.CharField(max_length=10, null=True, blank=True)
    collapse_arguments = models.BooleanField(
        default=True,
        help_text=_('Collapse separate inputs to a given argument to a single input (ie: --arg 1 --arg 2 becomes --arg 1 2)')
    )
    form_field = models.CharField(max_length=255)
    default = JSONCharField(max_length=255, null=True, blank=True)
    input_type = models.CharField(
        max_length=255,
        help_text=_('The python type expected by the script (e.g. boolean, integer, file).'),
    )
    custom_widget = models.ForeignKey('WooeyWidget', null=True, blank=True, on_delete=models.SET_NULL)
    param_help = models.TextField(verbose_name=_('help'), null=True, blank=True)
    is_checked = models.BooleanField(default=False)
    hidden = models.BooleanField(default=False)
    parameter_group = models.ForeignKey('ScriptParameterGroup', on_delete=models.PROTECT)
    param_order = models.SmallIntegerField(help_text=_('The order the parameter appears to the user.'), default=0)

    class Meta:
        app_label = 'wooey'
        verbose_name = _('script parameter')
        verbose_name_plural = _('script parameters')

    @property
    def form_slug(self):
        return '{}-{}'.format(self.parser.pk, self.slug)

    @property
    def multiple_choice(self):
        choice_limit = json.loads(self.choice_limit)
        if choice_limit is None:
            return False
        try:
            choice_limit = int(choice_limit)
        except ValueError:
            # it's not a set # of choices that is a max, it's either >=0, or >=1, which are the same for a front-end
            # since validation of >=0 or >=1 is performed outside of the form.
            return True
        else:
            return choice_limit > 1

    @property
    def max_choices(self):
        choice_limit = json.loads(self.choice_limit)
        if choice_limit is None:
            return 1
        try:
            choice_limit = int(choice_limit)
        except ValueError:
            # for this, it's either >=0 or >=1 so as many as they want.
            return -1
        else:
            return choice_limit

    def __str__(self):
        scripts = ', '.join([i.script.script_name for i in self.script_version.all()])
        return '{}: {}'.format(scripts, self.script_param)


# TODO: find a better name for this class. Job parameter? SelectedParameter?
class ScriptParameters(WooeyPy2Mixin, models.Model):
    """
        This holds the actual parameters sent with the submission
    """
    # the details of the actual executed scripts
    job = models.ForeignKey('WooeyJob', on_delete=models.CASCADE)
    parameter = models.ForeignKey('ScriptParameter', on_delete=models.PROTECT)
    # we store a JSON dumped string in here to attempt to keep our types in order
    _value = models.TextField(db_column='value')

    BOOLEAN = 'BooleanField'
    CHAR = 'CharField'
    CHOICE = 'ChoiceField'
    FILE = 'FileField'
    FLOAT = 'FloatField'
    INTEGER = 'IntegerField'

    WOOEY_FIELD_MAP = {
        BOOLEAN: lambda x: str(x).lower() == 'true',
        CHAR: str,
        CHOICE: str,
        FLOAT: float,
        INTEGER: int,
    }

    class Meta:
        app_label = 'wooey'
        verbose_name = _('script parameters')

    def __str__(self):
        try:
            value = self.value
        except IOError:
            value = _('FILE NOT FOUND')
        except SuspiciousFileOperation:
            value = _('File outside of project')
        return '{}: {}'.format(self.parameter.script_param, value)

    def get_subprocess_value(self):
        value = self.value
        if self.value is None:
            return None
        field = self.parameter.form_field
        param = self.parameter.short_param
        com = {'parameter': param, 'script_parameter': self.parameter}
        if field == self.BOOLEAN:
            if value:
                return com
            else:
                del com['parameter']
        if field == self.FILE:
            if self.parameter.is_output:
                try:
                    value = value.path
                except AttributeError:
                    value = utils.get_storage(local=True).path(value)
                    # trim the output path, we don't want to be adding our platform specific paths to the output
                    op = self.job.get_output_path()
                    #TODO : use os.path.sep
                    value = value[value.find(op)+len(op)+1:]
            else:
                # make sure we have it locally otherwise download it
                if not utils.get_storage(local=True).exists(value.path):
                    new_path = utils.get_storage(local=True).save(value.path, value)
                    value = new_path
                else:
                    # return the string for processing
                    value = value.path
        try:
            float(value)
            value = str(value)
        except ValueError:
            pass
        com['value'] = value if isinstance(value, six.string_types) else six.u(value)
        return com

    def force_value(self, value):
        self._value = json.dumps(value)

    def recreate(self):
        # we want to change filefields to reflect whatever is the current job's path. This is currently used for
        # job resubmission
        value = json.loads(self._value)
        if value is not None:
            field = self.parameter.form_field
            if field == self.FILE:
                # we are perfectly fine using old input files instead of recreating them, so only check output files
                if self.parameter.is_output:
                    new_path = self.job.get_output_path()
                    new_root, new_id = os.path.split(new_path)
                    # we want to remove the root + the old job's pk
                    value = value[value.find(new_root)+len(new_root)+1:]
                    value = value[value.find(os.path.sep)+1:]
                    # we want to create a new path for the current job
                    path = os.path.join(new_path, self.parameter.slug if not value else value)
                    value = path
                    self._value = json.dumps(value)

    @property
    def value(self):
        value = json.loads(self._value)
        if value is not None:
            field = self.parameter.form_field
            if field == self.FILE:
                try:
                    with utils.get_storage_object(value, close=False) as value:
                        pass
                except IOError:
                    # this can occur when the storage object is not yet made for output
                    if self.parameter.is_output:
                        return value
                    raise IOError
        return value

    @value.setter
    def value(self, value):
        # coerce the value to the proper type and store as json to make it persistent as well as have json
        #  handle type conversion on the way back out
        field = self.parameter.form_field
        add_file = False
        checksum = None
        if field == self.CHAR:
            if value is None:
                value = None
            elif field == self.CHAR:
                if not value:
                    value = None
            else:
                value = self.WOOEY_FIELD_MAP[field](value)
        elif field == self.INTEGER:
            value = self.WOOEY_FIELD_MAP[field](value) if isinstance(value, int) or str(value).isdigit() else None
        elif field == self.BOOLEAN:
            if value is None or value == False:
                value = None
            if value:
                value = True
        elif field == self.FILE:
            if self.parameter.is_output:
                # make a fake object for it
                path = os.path.join(self.job.get_output_path(), self.parameter.slug if not value else value)
                value = path
            else:
                if value:
                    local_storage = utils.get_storage(local=True)
                    current_path = local_storage.path(value.name)
                    checksum = utils.get_checksum(path=value)
                    path = utils.get_upload_path(current_path, checksum=checksum)
                    if hasattr(value, 'size'):
                        filesize = value.size
                    elif issubclass(type(value), IOBase):
                        value.seek(0, 2)
                        filesize = value.tell()
                        value.seek(0)
                    else:
                        filesize = None
                    if not local_storage.exists(path) or (filesize is not None and local_storage.size(path) != filesize):
                        local_path = local_storage.save(path, value)
                    else:
                        local_path = local_storage.path(path)
                        local_path = os.path.join(os.path.split(path)[0], os.path.split(local_path)[1])
                    remote_storage = utils.get_storage(local=False)
                    if not remote_storage.exists(path) or (filesize is not None and remote_storage.size(path) != filesize):
                        local_path = remote_storage.save(local_path, value)
                    add_file = True
                    value = local_path
        self._value = json.dumps(value)
        if add_file:
            # make a WooeyFile so the user can share it/etc.
            # get the system path for the file
            local_path = utils.get_storage(local=True).path(local_path)
            fileinfo = utils.get_file_info(local_path)
            # save ourself first, we have to do this because we are referenced in WooeyFile
            self.save()
            if checksum is None:
                checksum = utils.get_checksum(path=local_path)
            wooey_file, file_created = WooeyFile.objects.get_or_create(checksum=checksum)
            if file_created:
                wooey_file.filetype = fileinfo.get('type')
                wooey_file.filepreview = fileinfo.get('preview')
                save_file = utils.get_storage().open(local_path)
                save_path = path
                wooey_file.filepath.save(save_path, save_file, save=False)
                wooey_file.filepath.name = save_path
                wooey_file.save()

            UserFile.objects.get_or_create(job=self.job, system_file=wooey_file,
                                           parameter=self, filename=os.path.split(local_path)[1])


class UserFile(WooeyPy2Mixin, models.Model):
    filename = models.TextField()
    job = models.ForeignKey('WooeyJob', on_delete=models.CASCADE)
    system_file = models.ForeignKey('WooeyFile', on_delete=models.CASCADE)
    parameter = models.ForeignKey('ScriptParameters', null=True, blank=True, on_delete=models.CASCADE)

    class Meta:
        app_label = 'wooey'

    def __str__(self):
        return '{}: {}'.format(self.job.job_name, self.system_file)


class WooeyFile(WooeyPy2Mixin, models.Model):
    filepath = models.FileField(max_length=500)
    filepreview = models.TextField(null=True, blank=True)
    filetype = models.CharField(max_length=255, null=True, blank=True)
    size_bytes = models.IntegerField(null=True)
    checksum = models.CharField(max_length=40, blank=True)

    class Meta:
        app_label = 'wooey'
        verbose_name = _('wooey file')
        verbose_name_plural = _('wooey files')

    def __str__(self):
        return self.filepath.name



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-03-05 16:09
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('wooey', '0038_wooeyjob-choices'),
    ]

    operations = [
        migrations.CreateModel(
            name='WooeyWidget',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50, verbose_name='Widget Name')),
                ('input_attributes', models.TextField(blank=True, help_text='Extra attributes to the input field. The extra attributes MUST be specified like key="value".', null=True, verbose_name='Input Widget Extra Attributes')),
                ('input_class', models.CharField(blank=True, help_text='The class name(s) for the input field.', max_length=255, null=True, verbose_name='Input Widget Class name(s)')),
                ('input_properties', models.CharField(blank=True, help_text='Additional properties to append to the input field.', max_length=255, null=True, verbose_name='Input Widget Extra Properties')),
            ],
        ),
        migrations.AlterField(
            model_name='scriptparameter',
            name='param_order',
            field=models.SmallIntegerField(default=0, help_text='The order the parameter appears to the user.'),
        ),
        migrations.AddField(
            model_name='scriptparameter',
            name='custom_widget',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='wooey.WooeyWidget'),
        ),
    ]



from __future__ import absolute_import
__author__ = 'chris'
import copy
import json
import os
from collections import OrderedDict
from functools import partial

import six
from django import forms
from django.forms.utils import flatatt, format_html
from django.http.request import QueryDict
from django.utils.module_loading import import_string
from django.utils.safestring import mark_safe

from . import config
from .scripts import WooeyForm
from .. import version
from ..backend import utils
from ..models import ScriptVersion


def mutli_render(render_func, appender_data_dict=None):
    def render(name, value=None, attrs=None, renderer=None):
        if not isinstance(value, (list, tuple)):
            value = [value]
        values = value
        widget_renderer = partial(render_func, renderer=renderer) if renderer else partial(render_func)
        # The tag is a marker for our javascript to reshuffle the elements. This is because some widgets have complex rendering with multiple fields
        pieces = ['<{tag} {multi_attr}>{widget}</{tag}>'.format(tag='div', multi_attr=config.WOOEY_MULTI_WIDGET_ATTR,
                                                                widget=widget_renderer(name, value, attrs)) for value in values]

        # we add a final piece that is our button to click for adding. It's useful to have it here instead of the template so we don't
        # have to reverse-engineer who goes with what

        # build the attribute dict
        data_attrs = flatatt(appender_data_dict if appender_data_dict is not None else {})
        pieces.append(format_html('<a href="#{anchor}"{data}><span class="glyphicon glyphicon-plus"></span></a>', anchor=config.WOOEY_MULTI_WIDGET_ANCHOR
                                  , data=data_attrs))
        return mark_safe('\n'.join(pieces))
    return render


def multi_value_from_datadict(func):
    def value_from_datadict(data, files, name):
        return [func(QueryDict('{name}={value}'.format(name=name, value=six.moves.urllib.parse.quote(i))), files, name) for i in data.getlist(name)]
    return value_from_datadict


def multi_value_clean(func):
    def clean(*args, **kwargs):
        args = list(args)
        values = args[0]
        ret = []
        for value in values:
            value_args = args
            value_args[0] = value
            ret.append(func(*value_args, **kwargs))
        return ret
    return clean


class WooeyFormFactory(object):
    wooey_forms = {}

    @staticmethod
    def get_field(param, initial=None):
        """
        Any extra field attributes for the widget for customization of Wooey at the field level
         can be added to the widget dictionary, widget_data_dict, or to the appender_data_dict, which
         is the little plus button. This is useful since there is only a single copy of the plus,
         whereas we can have multiple widgets. Thus, javascript attributes we want to add per parameter
         can be added to appender_data_dict, and attributes we want to add to the widget input itself
         can be added to the widget_data_dict.

        :return: a field class
        """
        form_field = param.form_field
        widget_data_dict = {}
        appender_data_dict = {}
        widget_init = {}
        WOOEY_CHOICE_LIMIT = 'data-wooey-choice-limit'
        choices = json.loads(param.choices)
        field_kwargs = {
            'label': param.script_param.replace('_', ' ').title(),
            'required': param.required,
            'help_text': param.param_help,
        }
        multiple_choices = param.multiple_choice
        choice_limit = param.max_choices
        if initial is None and param.default is not None:
            initial = param.default
        if choices:
            form_field = 'MultipleChoiceField' if multiple_choices else 'ChoiceField'
            base_choices = [(None, '----')] if not param.required and not multiple_choices else []
            field_kwargs['choices'] = base_choices+[(str(i), str(i).title()) for i in choices]

        custom_widget = param.custom_widget
        if custom_widget:
            if custom_widget.widget_class:
                widget_class = import_string(custom_widget.widget_class)
                field_kwargs['widget'] = widget_class
            widget_init['attrs'] = custom_widget.widget_attributes

        if form_field == 'FileField':
            if param.is_output:
                form_field = 'CharField'
                if initial:
                    if not isinstance(initial, (list, tuple)):
                        initial = [initial]
                    initial = [os.path.split(i.name)[1] for i in initial]
            elif initial is not None and list(filter(None, initial)):  # for python3, we need to evaluate the filter object
                if isinstance(initial, (list, tuple)):
                    _initial = []
                    for value in initial:
                        if not hasattr(value, 'path'):
                            with utils.get_storage_object(value, close=False) as so:
                                _initial.append(so)
                        else:
                            _initial.append(value)
                    initial = _initial
                else:
                    if not hasattr(initial, 'path'):
                        with utils.get_storage_object(initial, close=False) as so:
                            initial = so
                    else:
                        initial = initial
                if not field_kwargs.get('widget'):
                    field_kwargs['widget'] = forms.ClearableFileInput
        if not multiple_choices and isinstance(initial, list):
            initial = initial[0]
        field_kwargs['initial'] = initial
        field = getattr(forms, form_field)

        if 'widget' in field_kwargs:
            field_kwargs['widget'] = field_kwargs['widget'](**widget_init)

        field = field(**field_kwargs)

        if form_field != 'MultipleChoiceField' and multiple_choices:
            field.widget.render = mutli_render(field.widget.render, appender_data_dict=appender_data_dict)
            field.widget.value_from_datadict = multi_value_from_datadict(field.widget.value_from_datadict)
            field.clean = multi_value_clean(field.clean)
            if choice_limit > 0:
                appender_data_dict[WOOEY_CHOICE_LIMIT] = choice_limit
        elif multiple_choices and choice_limit>0:
            widget_data_dict[WOOEY_CHOICE_LIMIT] = choice_limit

        field.widget.attrs.update(widget_data_dict)
        return field

    def get_group_forms(self, script_version=None, initial_dict=None, render_fn=None):
        REQUIRED_GROUP = 'Required'
        if initial_dict is None:
            initial_dict = {}
        params = [i for i in script_version.get_parameters() if not i.hidden]
        base_group_map = OrderedDict({
            REQUIRED_GROUP: {
                'group': REQUIRED_GROUP,
                'fields': OrderedDict(),
            }
        })
        parser_group_map = OrderedDict()

        for param in params:
            if param.parameter_group.hidden:
                continue
            parser = (param.parser_id, param.parser.name or '')
            group_map = parser_group_map.setdefault(parser, copy.deepcopy(base_group_map))
            initial_values = initial_dict.get(param.form_slug, None)
            field = self.get_field(param, initial=initial_values)
            field.name = param.form_slug
            group_name = REQUIRED_GROUP if param.required else param.parameter_group.group_name
            group = group_map.get(group_name, {
                'group': group_name,
                'fields': OrderedDict()
            })
            group['fields'][param.form_slug] = field
            group_map[group_name] = group

        # Add any missing parsers. This is needed if the script has no parameters
        added_subparsers = {i[1] for i in parser_group_map.keys()}
        for parser in script_version.scriptparser_set.order_by('pk'):
            if parser.name not in added_subparsers:
                parser_group_map[(parser.id, parser.name)] = copy.deepcopy(base_group_map)

        # If there are no required groups in a parser, remove them
        for parser, group_map in six.iteritems(parser_group_map):
            if not len(group_map[REQUIRED_GROUP]['fields']):
                del group_map[REQUIRED_GROUP]

        pk = script_version.pk
        form_initial = {
            'wooey_type': pk,
        }
        if initial_dict.get('wooey_parser') is not None:
            form_initial['wooey_parser'] = initial_dict['wooey_parser']
        wooey_form = WooeyForm(initial=form_initial)
        script_info = {
            'action': script_version.get_url(),
            'parsers': OrderedDict(),
            'wooey_form': wooey_form,
        }

        # create individual forms for each group and subparser
        for parser, group_map in six.iteritems(parser_group_map):
            parser_pk = parser[0]
            if wooey_form.fields['wooey_parser'].initial is None and parser_pk is not None:
                wooey_form.fields['wooey_parser'].initial = parser_pk
            parser_groups = script_info['parsers'].setdefault(parser, [])
            for group_index, group in enumerate(six.iteritems(group_map)):
                group_pk, group_info = group
                form = forms.Form()
                for form_slug, field in six.iteritems(group_info['fields']):
                    form.fields[form_slug] = field

                if render_fn:
                    form = render_fn(form)

                parser_groups.append({'group_name': group_info['group'], 'form': form})

        return script_info

    def get_master_form(self, script_version=None, pk=None, parser=None):
        pk = int(pk) if pk is not None else pk

        if script_version is None and pk is not None:
            script_version = ScriptVersion.objects.get(pk=pk)
        pk = script_version.pk

        master_form = WooeyForm(initial={'wooey_type': pk})

        params = script_version.get_parameters()
        for param in params:
            parser_name = param.parser.name
            if not parser or (parser_name == '' or param.parser_id == parser):
                field = self.get_field(param)
                master_form.fields[param.form_slug] = field

        return master_form

DJ_FORM_FACTORY = WooeyFormFactory()



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-07-02 14:31
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('wooey', '0025_remove_script_parameter_sv'),
    ]

    operations = [
        migrations.RenameField(
            model_name='scriptparameter',
            old_name='script_versions',
            new_name='script_version',
        ),
    ]



import os


from django.test import Client, TestCase, TransactionTestCase
from six.moves.urllib_parse import quote

from wooey import models, version

from . import factories, config, mixins, utils as test_utils


class ScriptTestCase(mixins.ScriptFactoryMixin, TestCase):

    def test_multiple_choices(self):
        multiple_choice_param = 'two_choices'
        single_choice_param = 'one_choice'
        optional_choice_param = 'all_choices'
        # test that we are a multiple choice entry
        param = models.ScriptParameter.objects.get(slug=multiple_choice_param)
        self.assertTrue(param.multiple_choice)

        # test our limit
        self.assertEqual(param.max_choices, 2)

        # test with a singular case
        param = models.ScriptParameter.objects.get(slug=single_choice_param)
        self.assertFalse(param.multiple_choice)
        self.assertEqual(param.max_choices, 1)

        # test cases that have variable requirements
        param = models.ScriptParameter.objects.get(slug=optional_choice_param)
        self.assertTrue(param.multiple_choice)
        self.assertEqual(param.max_choices, -1)

    def test_deletes_related_objects(self):
        self.assertTrue(models.ScriptVersion.objects.filter(pk=self.choice_script.pk).exists())
        script = models.Script.objects.get(pk=self.choice_script.script.pk)
        script.delete()
        self.assertFalse(models.ScriptVersion.objects.filter(pk=self.choice_script.pk).exists())



class ScriptGroupTestCase(TestCase):

    def test_script_group_creation(self):
        group = factories.ScriptGroupFactory()

class TestScriptParsers(mixins.ScriptFactoryMixin, TestCase):
    def test_renders_if_script_version_deleted(self):
        parser = self.choice_script.scriptparser_set.first()
        self.choice_script.delete()
        self.assertIn(parser.name, str(parser))

class ScriptParameterTestCase(TestCase):
    def test_script_parameter_default(self):
        script_parameter = factories.ScriptParameterFactory()
        pk = script_parameter.pk
        for test_value in [123, 'abc', {'abc': 5}]:
            script_parameter.default = test_value
            script_parameter.save()
            self.assertEqual(models.ScriptParameter.objects.get(pk=pk).default, test_value)

class TestScriptVersion(mixins.ScriptFactoryMixin, TestCase):
    def test_script_version_url_with_spaces(self):
        # Handles https://github.com/wooey/Wooey/issues/290
        script_version = self.choice_script
        spaced_version = 'v 1 0 0'
        script_version.script_version = spaced_version
        script_version.save()
        url = script_version.get_version_url()
        self.assertIn(quote(spaced_version), url)

class TestJob(mixins.ScriptFactoryMixin, mixins.FileCleanupMixin, mixins.FileMixin, TransactionTestCase):

    def get_local_url(self, fileinfo):
        from ..backend import utils
        local_storage = utils.get_storage(local=True)
        return local_storage.url(fileinfo['object'].filepath.name)

    def test_jobs(self):
        script = self.translate_script
        from ..backend import utils
        sequence_slug = test_utils.get_subparser_form_slug(script, 'sequence')
        out_slug = test_utils.get_subparser_form_slug(script, 'out')
        fasta_slug = test_utils.get_subparser_form_slug(script, 'fasta')
        job = utils.create_wooey_job(
            script_version_pk=script.pk,
            data={
                'job_name': 'abc',
                sequence_slug: 'aaa',
                out_slug: 'abc'
            }
        )
        job = job.submit_to_celery()
        old_pk = job.pk
        new_job = job.submit_to_celery(resubmit=True)
        self.assertNotEqual(old_pk, new_job.pk)
        # test rerunning, our output should be removed
        from ..models import UserFile
        old_output = sorted([i.pk for i in UserFile.objects.filter(job=new_job)])
        # the pk will not change here since we are using rerun=True
        new_job.submit_to_celery(rerun=True)
        # check that we overwrite our output
        new_output = sorted([i.pk for i in UserFile.objects.filter(job=new_job)])
        self.assertNotEqual(old_output, new_output)
        self.assertEqual(len(old_output), len(new_output))
        # check the old entries are gone
        self.assertEqual([], list(UserFile.objects.filter(pk__in=old_output)))

        file_previews = utils.get_file_previews(job)
        for group, files in file_previews.items():
            for fileinfo in files:
                # for testing, we use the local url
                response = Client().get(self.get_local_url(fileinfo))
                self.assertEqual(response.status_code, 200)

        # check our download links are ok
        # upload the file first to our storage engine so this works in tests
        local_storage = utils.get_storage(local=True)
        fasta_path = local_storage.save('fasta.fasta', open(os.path.join(config.WOOEY_TEST_DATA, 'fasta.fasta')))
        fasta_file = local_storage.open(fasta_path)
        job = utils.create_wooey_job(
            script_version_pk=script.pk,
            data={
                fasta_slug: fasta_file,
                out_slug: 'abc',
                'job_name': 'abc'
            }
        )

        # check our upload link is ok
        file_previews = utils.get_file_previews(job)
        for group, files in file_previews.items():
            for fileinfo in files:
                response = Client().get(self.get_local_url(fileinfo))
                self.assertEqual(response.status_code, 200)

    def test_file_sharing(self):
        # this tests whether a file uploaded by one job will be referenced by a second job instead of being duplicated
        # on the file system
        new_file = self.storage.open(self.get_any_file())
        script = self.choice_script
        script_slug = test_utils.get_subparser_form_slug(script, 'multiple_file_choices')
        from ..backend import utils
        job = utils.create_wooey_job(script_version_pk=script.pk, data={'job_name': 'job1', script_slug: new_file})
        job = job.submit_to_celery()
        job2 = utils.create_wooey_job(script_version_pk=script.pk, data={'job_name': 'job2', script_slug: new_file})
        job2 = job2.submit_to_celery()
        job1_files = [i for i in models.UserFile.objects.filter(job=job, parameter__isnull=False) if i.parameter.parameter.form_slug == script_slug]
        job1_file = job1_files[0]
        job2_files = [i for i in models.UserFile.objects.filter(job=job2, parameter__isnull=False) if i.parameter.parameter.form_slug == script_slug]
        job2_file = job2_files[0]
        self.assertNotEqual(job1_file.pk, job2_file.pk)
        self.assertEqual(job1_file.system_file, job2_file.system_file)


    def test_multiplechoices(self):
        script = self.choice_script
        choices = [2, 1, 3]
        choice_slug = test_utils.get_subparser_form_slug(script, 'two_choices')

        from ..backend import utils
        job = utils.create_wooey_job(
            script_version_pk=script.pk,
            data={
                'job_name': 'abc',
                choice_slug: choices
            }
        )
        # make sure we have our choices in the parameters
        choice_params = [i.value for i in job.get_parameters() if i.parameter.form_slug == choice_slug]
        self.assertEqual(choices, choice_params)
        job = job.submit_to_celery()


class TestCustomWidgets(TestCase):
    def test_widget_attributes(self):
        widget = factories.WooeyWidgetFactory(
            input_properties='custom-property',
            input_attributes='attr1="custom1" attr2="custom2"',
            input_class='custom-class',
        )
        self.assertEquals(
            widget.widget_attributes,
            {
                'custom-property': True,
                'attr1': 'custom1',
                'attr2': 'custom2',
                'class': 'custom-class',
            }
        )


__author__ = 'chris'
import traceback
import sys

from wooey.version import DJANGO_VERSION, DJ110
if DJANGO_VERSION >= DJ110:
    from django.utils.deprecation import MiddlewareMixin
else:
    MiddlewareMixin = object

class ProcessExceptionMiddleware(MiddlewareMixin):
    def process_response(self, request, response):
        if response.status_code != 200:
            try:
                sys.stderr.write('{}'.format(''.join(traceback.format_exc())))
            except AttributeError:
                pass
        return response


# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-07-03 09:48
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wooey', '0026_rename_script_parameter_sv'),
    ]

    operations = [
        migrations.AddField(
            model_name='scriptparameter',
            name='param_order',
            field=models.SmallIntegerField(default=0, verbose_name='The order the parameter appears to the user.'),
        ),
        migrations.AlterField(
            model_name='scriptparameter',
            name='script_version',
            field=models.ManyToManyField(to='wooey.ScriptVersion'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2017-04-25 09:25
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import wooey.models.mixins


class Migration(migrations.Migration):

    dependencies = [
        ('wooey', '0027_parameter_order'),
    ]

    operations = [
        migrations.CreateModel(
            name='ScriptParser',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, max_length=255, default='')),
                ('script_version', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wooey.ScriptVersion')),
            ],
            bases=(wooey.models.mixins.WooeyPy2Mixin, models.Model),
        ),
        migrations.AddField(
            model_name='scriptparameter',
            name='parser',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='wooey.ScriptParser'),
            preserve_default=False,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import autoslug.fields


class Migration(migrations.Migration):

    dependencies = [
        ('wooey', '0002_remove_scriptparameter_output_path'),
    ]

    operations = [
        migrations.AlterField(
            model_name='script',
            name='slug',
            field=autoslug.fields.AutoSlugField(populate_from='script_name', unique=True, editable=False),
        ),
        migrations.AlterField(
            model_name='scriptgroup',
            name='slug',
            field=autoslug.fields.AutoSlugField(populate_from='group_name', unique=True, editable=False),
        ),
        migrations.AlterField(
            model_name='scriptparameter',
            name='slug',
            field=autoslug.fields.AutoSlugField(populate_from='script_param', unique=True, editable=False),
        ),
    ]
